@@ -3,28 +3,20 @@
 # specifically for storing facts (Predicate instances).
 # ------------------------------------------------------------------------------
 
-import sys
-import io
-import operator
-import collections
-import bisect
-import inspect
 import abc
-import functools
+import io
 import itertools
+import sys
+from typing import (Any, Callable, Iterable, Iterator, List, Optional, TextIO,
+                    Tuple, Type, Union, cast, overload)
 
-from .core import *
-from .core import get_field_definition, PredicatePath, kwargs_check_keys, \
-    validate_root_paths
-
-from .query import *
-
-from .query import Placeholder, OrderBy, desc, asc
-
-from .query import process_where, process_join, process_orderby, \
-    make_query_plan, QuerySpec, QueryExecutor
-
-from .factcontainers import FactSet, FactIndex, FactMap, factset_equality
+from ._typing import _T0, _T1, _T2, _T3, _T4
+from ._queryimpl import UnGroupedQuery
+from .core import (Predicate, PredicateDefn, PredicatePath, and_,
+                   validate_root_paths)
+from .factcontainers import FactMap, factset_equality
+from .query import (QueryExecutor, QuerySpec, make_query_plan, process_orderby,
+                    process_where)
 
 __all__ = [
     'FactBase',
@@ -35,6 +27,7 @@ __all__ = [
 #------------------------------------------------------------------------------
 # Global
 #------------------------------------------------------------------------------
+_Facts = Union[Iterable[Predicate], Callable[[], Iterable[Predicate]]]
 
 #------------------------------------------------------------------------------
 # Support function for printing ASP facts: Note: _trim_docstring() is taken from
@@ -96,17 +89,23 @@ def _format_docstring(docstring,output):
 def _maxwidth(lines):
     return max([len(l) for l in lines])
 
-def _format_commented(fm,out):
-    pm = fm.predicate.meta
+def _format_commented(fm: FactMap, out: TextIO) -> None:
+    pm: PredicateDefn = fm.predicate.meta
     docstring = _trim_docstring(fm.predicate.__doc__) \
         if fm.predicate.__doc__ else ""
     indent = "    "
     if pm.arity == 0:
         lines = [ "Unary predicate signature:", indent + pm.name ]
     else:
-        params=[str(fa.name) for fa in pm]
+        def build_signature(p: Type[Predicate]) -> str:
+            args = []
+            for pp in p:
+                complex = pp.meta.field.complex
+                args.append(cast(str, pp._pathseq[1]) if not complex else build_signature(complex))
+            return f"{p.meta.name}({','.join(args)})"
+
         lines = [ "Predicate signature:",
-                  indent + "{}({})".format(pm.name, ",".join(params)) ]
+                  indent + build_signature(fm.predicate) ]
     if docstring:
         lines.append("Description:")
         for l in docstring.splitlines():lines.append(indent + l)
@@ -162,7 +161,7 @@ class FactBase(object):
     def _init(self, facts=None, indexes=None):
 
         # flag that initialisation has taken place
-        self._delayed_init = None
+        self._delayed_init: Optional[Callable[[], None]] = None
 
         # If it is delayed initialisation then get the facts
         if facts and callable(facts):
@@ -191,25 +190,24 @@ class FactBase(object):
     #
     #--------------------------------------------------------------------------
 
-    def _add(self, arg):
+    def _add(self, arg: Union[Predicate, Iterable[Predicate]]) -> None:
         if isinstance(arg, Predicate):
-            return self._add_fact(type(arg),arg)
-        elif isinstance(arg, str) or not isinstance(arg, collections.abc.Iterable):
+            ptype = arg.__class__
+            if not ptype  in self._factmaps:
+                self._factmaps[ptype] = FactMap(ptype)
+            return self._factmaps[ptype].add_fact(arg)
+
+        if isinstance(arg, str) or not isinstance(arg, Iterable):
             raise TypeError(f"'{arg}' is not a Predicate instance")
 
-        facts = sorted(arg, key=lambda x : type(x).__name__)
-        for ptype, g in itertools.groupby(facts, lambda x: type(x)):
-            self._add_facts(ptype, g)
-
-    def _add_fact(self, ptype, fact):
-        fm = self._factmaps.setdefault(ptype, FactMap(ptype))
-        fm.add_fact(fact)
-
-    def _add_facts(self, ptype, facts):
-        if not issubclass(ptype,Predicate):
-            raise TypeError(f"{list(facts)} are not Predicate instances")
-        fm = self._factmaps.setdefault(ptype, FactMap(ptype))
-        fm.add_facts(facts)
+        sorted_facts = sorted(arg, key=lambda x: x.__class__.__name__)
+        for type_, grouped_facts in itertools.groupby(sorted_facts, lambda x: x.__class__):
+            if not issubclass(type_, Predicate):
+                raise TypeError(f"{list(grouped_facts)} are not Predicate instances")
+            if not type_ in self._factmaps:
+                self._factmaps[type_] = FactMap(type_)
+            self._factmaps[type_].add_facts(grouped_facts)
+        return
 
     def _remove(self, fact, raise_on_missing):
         ptype = type(fact)
@@ -221,7 +219,7 @@ class FactBase(object):
     #--------------------------------------------------------------------------
     # Initiliser
     #--------------------------------------------------------------------------
-    def __init__(self, facts=None, indexes=None):
+    def __init__(self, facts: Optional[_Facts] = None, indexes: Optional[Iterable[PredicatePath]] = None) -> None:
         self._delayed_init=None
         if callable(facts):
             def delayed_init():
@@ -242,7 +240,7 @@ class FactBase(object):
     #--------------------------------------------------------------------------
     # Set member functions
     #--------------------------------------------------------------------------
-    def add(self, arg):
+    def add(self, arg: Union[Predicate, Iterable[Predicate]]) -> None:
         """Add a single fact or a collection of facts.
 
         Because a ``FactBase`` can only hold :class:`~Predicate` sub-class
@@ -257,17 +255,17 @@ class FactBase(object):
         self._check_init()  # Check for delayed init
         return self._add(arg)
 
-    def remove(self, arg):
+    def remove(self, arg: Predicate) -> None:
         """Remove a fact from the fact base (raises an exception if no fact). """
         self._check_init()  # Check for delayed init
         return self._remove(arg, raise_on_missing=True)
 
-    def discard(self, arg):
+    def discard(self, arg: Predicate) -> None:
         """Remove a fact from the fact base. """
         self._check_init()  # Check for delayed init
         return self._remove(arg, raise_on_missing=False)
 
-    def pop(self):
+    def pop(self) -> Predicate:
         """Pop an element from the FactBase. """
         self._check_init()  # Check for delayed init
         for pt, fm in self._factmaps.items():
@@ -320,6 +318,46 @@ class FactBase(object):
 
         return _Delete(self, QuerySpec(roots=roots))
 
+    # START OVERLOADED FUNCTIONS self.query;UnGroupedQuery[{0}];1;5;Type;
+
+    # code within this block is **programmatically, 
+    # statically generated** by generate_overloads.py
+
+    @overload
+    def query(
+        self, __ent0: Type[_T0]
+    ) -> 'UnGroupedQuery[_T0]':
+        ...
+
+    @overload
+    def query(
+        self, __ent0: Type[_T0], __ent1: Type[_T1]
+    ) -> 'UnGroupedQuery[Tuple[_T0, _T1]]':
+        ...
+
+    @overload
+    def query(
+        self, __ent0: Type[_T0], __ent1: Type[_T1], __ent2: Type[_T2]
+    ) -> 'UnGroupedQuery[Tuple[_T0, _T1, _T2]]':
+        ...
+
+    @overload
+    def query(
+        self, __ent0: Type[_T0], __ent1: Type[_T1], __ent2: Type[_T2], __ent3: Type[_T3]
+    ) -> 'UnGroupedQuery[Tuple[_T0, _T1, _T2, _T3]]':
+        ...
+
+    @overload
+    def query(
+        self, __ent0: Type[_T0], __ent1: Type[_T1], __ent2: Type[_T2], __ent3: Type[_T3], __ent4: Type[_T4]
+    ) -> 'UnGroupedQuery[Tuple[_T0, _T1, _T2, _T3, _T4]]':
+        ...
+
+    # END OVERLOADED FUNCTIONS self.query
+   
+    @overload
+    def query(self, *roots: Any) -> 'UnGroupedQuery[Any]': ...
+
     def query(self, *roots):
         """Define a query using the new Query API :class:`Query`.
 
@@ -341,28 +379,28 @@ class FactBase(object):
         for ptype in ptypes: self._factmaps.setdefault(ptype, FactMap(ptype))
 
         qspec = QuerySpec(roots=roots)
-        return QueryImpl(self._factmaps, qspec)
+        return UnGroupedQuery(self._factmaps, qspec)
 
     @property
-    def predicates(self):
+    def predicates(self) -> Tuple[Type[Predicate], ...]:
         """Return the list of predicate types that this fact base contains."""
 
         self._check_init()  # Check for delayed init
         return tuple([pt for pt, fm in self._factmaps.items() if fm])
 
     @property
-    def indexes(self):
+    def indexes(self) -> Tuple[PredicatePath, ...]:
         self._check_init()  # Check for delayed init
         return self._indexes
 
-    def facts(self):
+    def facts(self) -> List[Predicate]:
         """Return all facts."""
 
         self._check_init()  # Check for delayed init
         tmp = [ fm.factset for fm in self._factmaps.values() if fm]
         return list(itertools.chain(*tmp))
 
-    def asp_str(self,*,width=0,commented=False,sorted=False):
+    def asp_str(self, *, width: int = 0, commented: bool = False, sorted: bool = False) -> str:
         """Return a ASP string representation of the fact base.
 
         The generated ASP string representation is syntactically correct ASP
@@ -412,9 +450,7 @@ class FactBase(object):
         out.close()
         return data
 
-
-
-    def __str__(self):
+    def __str__(self) -> str:
         self._check_init()  # Check for delayed init
         tmp = ", ".join([str(f) for f in self])
         return '{' + tmp + '}'
@@ -449,7 +485,7 @@ class FactBase(object):
         self._check_init() # Check for delayed init
         return sum([len(fm.factset) for fm in self._factmaps.values()])
 
-    def __iter__(self):
+    def __iter__(self) -> Iterator[Predicate]:
         self._check_init() # Check for delayed init
 
         for fm in self._factmaps.values():
@@ -472,12 +508,6 @@ class FactBase(object):
 
         return True
 
-    def __ne__(self, other):
-        """Overloaded boolean operator."""
-        result = self.__eq__(other)
-        if result is NotImplemented: return NotImplemented
-        return not result
-
     def __lt__(self,other):
         """Implemement set < operator."""
 
@@ -569,60 +599,60 @@ class FactBase(object):
     #--------------------------------------------------------------------------
     # Set functions
     #--------------------------------------------------------------------------
-    def union(self,*others):
+    def union(self, *others: _Facts) -> 'FactBase':
         """Implements the set union() function"""
-        others=[o if isinstance(o, self.__class__) else FactBase(o) for o in others]
+        factbases = [o if isinstance(o, self.__class__) else FactBase(o) for o in others]
         self._check_init() # Check for delayed init
-        for o in others: o._check_init()
+        for fb in factbases: fb._check_init()
 
-        fb=FactBase()
+        fb = FactBase()
         predicates = set(self._factmaps.keys())
-        for o in others: predicates.update(o._factmaps.keys())
+        for o in factbases: predicates.update(o._factmaps.keys())
 
         for p in predicates:
-            pothers=[ o._factmaps[p] for o in others if p in o._factmaps]
+            pothers = [o._factmaps[p] for o in factbases if p in o._factmaps]
             if p in self._factmaps:
                 fb._factmaps[p] = self._factmaps[p].union(*pothers)
             else:
                 fb._factmaps[p] = FactMap(p).union(*pothers)
         return fb
 
-    def intersection(self,*others):
+    def intersection(self, *others: _Facts) -> 'FactBase':
         """Implements the set intersection() function"""
-        others=[o if isinstance(o, self.__class__) else FactBase(o) for o in others]
+        factbases = [o if isinstance(o, self.__class__) else FactBase(o) for o in others]
         self._check_init() # Check for delayed init
-        for o in others: o._check_init()
+        for fb in factbases: fb._check_init()
 
-        fb=FactBase()
+        fb = FactBase()
         predicates = set(self._factmaps.keys())
-        for o in others: predicates.intersection_update(o._factmaps.keys())
+        for fb_ in factbases: predicates.intersection_update(fb_._factmaps.keys())
 
         for p in predicates:
-            pothers=[ o._factmaps[p] for o in others if p in o._factmaps]
+            pothers = [o._factmaps[p] for o in factbases if p in o._factmaps]
             fb._factmaps[p] = self._factmaps[p].intersection(*pothers)
         return fb
 
-    def difference(self,*others):
+    def difference(self, *others: _Facts) -> 'FactBase':
         """Implements the set difference() function"""
-        others=[o if isinstance(o, self.__class__) else FactBase(o) for o in others]
+        factbases = [o if isinstance(o, self.__class__) else FactBase(o) for o in others]
         self._check_init() # Check for delayed init
-        for o in others: o._check_init()
+        for fb in factbases: fb._check_init()
 
-        fb=FactBase()
+        fb = FactBase()
         predicates = set(self._factmaps.keys())
 
         for p in predicates:
-            pothers=[ o._factmaps[p] for o in others if p in o._factmaps]
+            pothers = [o._factmaps[p] for o in factbases if p in o._factmaps]
             fb._factmaps[p] = self._factmaps[p].difference(*pothers)
         return fb
 
-    def symmetric_difference(self,other):
+    def symmetric_difference(self, other: _Facts) -> 'FactBase':
         """Implements the set symmetric_difference() function"""
         if not isinstance(other, self.__class__): other=FactBase(other)
         self._check_init() # Check for delayed init
         other._check_init()
 
-        fb=FactBase()
+        fb = FactBase()
         predicates = set(self._factmaps.keys())
         predicates.update(other._factmaps.keys())
 
@@ -637,44 +667,43 @@ class FactBase(object):
 
         return fb
 
-    def update(self,*others):
+    def update(self, *others: _Facts) -> None:
         """Implements the set update() function"""
-        others=[o if isinstance(o, self.__class__) else FactBase(o) for o in others]
+        factbases = [o if isinstance(o, self.__class__) else FactBase(o) for o in others]
         self._check_init() # Check for delayed init
-        for o in others: o._check_init()
+        for fb in factbases: fb._check_init()
 
-        for o in others:
-            for p,fm in o._factmaps.items():
+        for fb in factbases:
+            for p,fm in fb._factmaps.items():
                 if p in self._factmaps: self._factmaps[p].update(fm)
                 else: self._factmaps[p] = fm.copy()
 
-    def intersection_update(self,*others):
+    def intersection_update(self, *others: _Facts) -> None:
         """Implements the set intersection_update() function"""
-        others=[o if isinstance(o, self.__class__) else FactBase(o) for o in others]
+        factbases = [o if isinstance(o, self.__class__) else FactBase(o) for o in others]
         self._check_init() # Check for delayed init
-        for o in others: o._check_init()
-        num = len(others)
+        for fb in factbases: fb._check_init()
 
         predicates = set(self._factmaps.keys())
-        for o in others: predicates.intersection_update(o._factmaps.keys())
+        for fb in factbases: predicates.intersection_update(fb._factmaps.keys())
         pred_to_delete = set(self._factmaps.keys()) - predicates
 
         for p in pred_to_delete: self._factmaps[p].clear()
         for p in predicates:
-            pothers=[ o._factmaps[p] for o in others if p in o._factmaps]
+            pothers = [o._factmaps[p] for o in factbases if p in o._factmaps]
             self._factmaps[p].intersection_update(*pothers)
 
-    def difference_update(self,*others):
+    def difference_update(self, *others: _Facts) -> None:
         """Implements the set difference_update() function"""
-        others=[o if isinstance(o, self.__class__) else FactBase(o) for o in others]
+        factbases = [o if isinstance(o, self.__class__) else FactBase(o) for o in others]
         self._check_init() # Check for delayed init
-        for o in others: o._check_init()
+        for fb in factbases: fb._check_init()
 
         for p in self._factmaps.keys():
-            pothers=[ o._factmaps[p] for o in others if p in o._factmaps ]
+            pothers = [o._factmaps[p] for o in factbases if p in o._factmaps]
             self._factmaps[p].difference_update(*pothers)
 
-    def symmetric_difference_update(self,other):
+    def symmetric_difference_update(self, other: _Facts) -> None:
         """Implements the set symmetric_difference_update() function"""
         if not isinstance(other, self.__class__): other=FactBase(other)
         self._check_init() # Check for delayed init
@@ -689,12 +718,11 @@ class FactBase(object):
             else:
                 if p in other._factmaps: self._factmaps[p] = other._factmaps[p].copy()
 
-
-    def copy(self):
+    def copy(self) -> 'FactBase':
         """Implements the set copy() function"""
         self._check_init() # Check for delayed init
-        fb=FactBase()
-        for p,fm in self._factmaps.items():
+        fb = FactBase()
+        for p, _ in self._factmaps.items():
             fb._factmaps[p] = self._factmaps[p].copy()
         return fb
 
@@ -941,222 +969,6 @@ class _Delete(Delete):
         for fact in to_delete: factmap.remove(fact)
         return len(to_delete)
 
-
-#------------------------------------------------------------------------------
-# New Clorm Query API
-#
-# QueryImpl
-# - factmaps             - dictionary mapping predicate types to FactMap objects
-# - qspec                - a dictionary with query parameters
-#------------------------------------------------------------------------------
-class QueryImpl(Query):
-
-    def __init__(self, factmaps, qspec):
-        self._factmaps = factmaps
-        self._qspec = qspec
-
-    #--------------------------------------------------------------------------
-    # Internal function to test whether a function has been called and add it
-    #--------------------------------------------------------------------------
-    def _check_join_called_first(self, name,endpoint=False):
-        if self._qspec.join is not None or len(self._qspec.roots) == 1: return
-        if endpoint:
-            raise ValueError(("A query over multiple predicates is incomplete without "
-                              "'join' clauses connecting these predicates"))
-        raise ValueError("A 'join' clause must be specified before '{}'".format(name))
-
-    #--------------------------------------------------------------------------
-    # Add a join expression
-    #--------------------------------------------------------------------------
-    def join(self, *expressions):
-        join=process_join(expressions, self._qspec.roots)
-        return QueryImpl(self._factmaps, self._qspec.newp(join=join))
-
-    #--------------------------------------------------------------------------
-    # Add an order_by expression
-    #--------------------------------------------------------------------------
-    def where(self, *expressions):
-        self._check_join_called_first("where")
-
-        if not expressions:
-            self._qspec.newp(where=None)    # Raise an error
-
-        if len(expressions) == 1:
-            where = process_where(expressions[0], self._qspec.roots)
-        else:
-            where = process_where(and_(*expressions), self._qspec.roots)
-
-        nqspec = self._qspec.newp(where=where)
-        return QueryImpl(self._factmaps, nqspec)
-
-    #--------------------------------------------------------------------------
-    # Add an orderered() flag
-    #--------------------------------------------------------------------------
-    def ordered(self, *expressions):
-        self._check_join_called_first("ordered")
-        if self._qspec.getp("order_by",None) is not None:
-            raise ValueError(("Invalid query 'ordered' declaration conflicts "
-                              "with previous 'order_by' declaration"))
-        nqspec = self._qspec.newp(ordered=True)
-        return QueryImpl(self._factmaps, nqspec)
-
-    #--------------------------------------------------------------------------
-    # Add an order_by expression
-    #--------------------------------------------------------------------------
-    def order_by(self, *expressions):
-        self._check_join_called_first("order_by")
-        if not expressions:
-            nqspec = self._qspec.newp(order_by=None)   # raise exception
-        elif self._qspec.getp("ordered",False):
-            raise ValueError(("Invalid query 'order_by' declaration '{}' "
-                              "conflicts with previous 'ordered' "
-                              "declaration").format(expressions))
-        else:
-            nqspec = self._qspec.newp(
-                order_by=process_orderby(expressions,self._qspec.roots))
-        return QueryImpl(self._factmaps, nqspec)
-
-    #--------------------------------------------------------------------------
-    # Add a group_by expression
-    #--------------------------------------------------------------------------
-    def group_by(self, *expressions):
-        if not expressions:
-            nqspec = self._qspec.newp(group_by=None)   # raise exception
-        else:
-            nqspec = self._qspec.newp(
-                group_by=process_orderby(expressions,self._qspec.roots))
-        return QueryImpl(self._factmaps, nqspec)
-
-    #--------------------------------------------------------------------------
-    # Explicitly select the elements to output or delete
-    #--------------------------------------------------------------------------
-    def select(self,*outsig):
-        self._check_join_called_first("select")
-        if not outsig:
-            raise ValueError("An empty 'select' signature is invalid")
-        nqspec = self._qspec.newp(select=outsig)
-        return QueryImpl(self._factmaps, nqspec)
-
-    #--------------------------------------------------------------------------
-    # The distinct flag
-    #--------------------------------------------------------------------------
-    def distinct(self):
-        self._check_join_called_first("distinct")
-        nqspec = self._qspec.newp(distinct=True)
-        return QueryImpl(self._factmaps, nqspec)
-
-    #--------------------------------------------------------------------------
-    # Ground - bind
-    #--------------------------------------------------------------------------
-    def bind(self,*args,**kwargs):
-        self._check_join_called_first("bind")
-        nqspec = self._qspec.bindp(*args, **kwargs)
-        return QueryImpl(self._factmaps, nqspec)
-
-    #--------------------------------------------------------------------------
-    # The tuple flag
-    #--------------------------------------------------------------------------
-    def tuple(self):
-        self._check_join_called_first("tuple")
-        nqspec = self._qspec.newp(tuple=True)
-        return QueryImpl(self._factmaps, nqspec)
-
-    #--------------------------------------------------------------------------
-    # Overide the default heuristic
-    #--------------------------------------------------------------------------
-    def heuristic(self, join_order):
-        nqspec = self._qspec.newp(heuristic=True, joh=join_order)
-        return QueryImpl(self._factmaps, nqspec)
-
-    #--------------------------------------------------------------------------
-    # End points that do something useful
-    #--------------------------------------------------------------------------
-
-    #--------------------------------------------------------------------------
-    # For the user to see what the query plan looks like
-    #--------------------------------------------------------------------------
-    def query_plan(self,*args,**kwargs):
-        self._check_join_called_first("query_plan")
-        qspec = self._qspec.fill_defaults()
-
-        (factsets,factindexes) = \
-            QueryExecutor.get_factmap_data(self._factmaps, qspec)
-        return make_query_plan(factindexes.keys(), qspec)
-
-    #--------------------------------------------------------------------------
-    # Return the placeholders
-    #--------------------------------------------------------------------------
-    @property
-    def qspec(self):
-        return self._qspec
-
-    #--------------------------------------------------------------------------
-    # Select to display all the output of the query
-    # --------------------------------------------------------------------------
-    def all(self):
-        self._check_join_called_first("all",endpoint=True)
-
-        qe = QueryExecutor(self._factmaps, self._qspec)
-        return qe.all()
-
-    #--------------------------------------------------------------------------
-    # Show the single element and throw an exception if there is more than one
-    # --------------------------------------------------------------------------
-    def singleton(self):
-        self._check_join_called_first("singleton",endpoint=True)
-
-        qe = QueryExecutor(self._factmaps, self._qspec)
-        count = 0
-        for out in qe.all():
-            count += 1
-            if count > 1:
-                raise ValueError("Query returned more than a single element")
-        if count == 0:
-            raise ValueError("Query has no matching elements")
-        return out
-
-    #--------------------------------------------------------------------------
-    # Return the count of elements - Note: the behaviour of what is counted
-    # changes if group_by() has been specified.
-    # --------------------------------------------------------------------------
-    def count(self):
-        self._check_join_called_first("count",endpoint=True)
-
-        qe = QueryExecutor(self._factmaps, self._qspec)
-
-        def group_by_generator():
-            for k, g in qe.all():
-                yield k, sum(1 for _ in g)
-
-        # If group_by is set then we want to count records associated with each
-        # key and not just total records.
-        if self._qspec.group_by:
-            return group_by_generator()
-        else:
-            return sum(1 for _ in qe.all())
-
-
-    #--------------------------------------------------------------------------
-    # Return the first element of the query
-    # --------------------------------------------------------------------------
-    def first(self):
-        self._check_join_called_first("first",endpoint=True)
-
-        qe = QueryExecutor(self._factmaps, self._qspec)
-
-        for out in qe.all():
-            return out
-        raise ValueError("Query has no matching elements")
-
-    #--------------------------------------------------------------------------
-    # Delete a selection of fact
-    #--------------------------------------------------------------------------
-    def delete(self):
-        self._check_join_called_first("delete",endpoint=True)
-
-        qe = QueryExecutor(self._factmaps, self._qspec)
-        return qe.delete()
-
 #------------------------------------------------------------------------------
 # main
 #------------------------------------------------------------------------------
